@@ -3,6 +3,7 @@
 # Copyright(c) 2015-2016: wave++ "Yuri D'Elia" <wavexx@thregr.org>
 # Copyright(c) 2019-2020: Yuto Tokunaga <yuntan.sub1@gmail.com>
 
+from . import inputlistener
 from .inputlistener import InputListener, InputType
 
 from gi.repository import GLib
@@ -54,22 +55,32 @@ REPLACE_SYMS = {
     'End':          KeyRepl(True,  True,  True,  _('End')),
     'Insert':       KeyRepl(False, True,  True,  _('Ins')),
     'Delete':       KeyRepl(True,  False, True,  _('Del')),
-    'KP_End':       KeyRepl(False, False, True,  _('(1)')),
-    'KP_Down':      KeyRepl(False, False, True,  _('(2)')),
-    'KP_Next':      KeyRepl(False, False, True,  _('(3)')),
-    'KP_Left':      KeyRepl(False, False, True,  _('(4)')),
-    'KP_Begin':     KeyRepl(False, False, True,  _('(5)')),
-    'KP_Right':     KeyRepl(False, False, True,  _('(6)')),
-    'KP_Home':      KeyRepl(False, False, True,  _('(7)')),
-    'KP_Up':        KeyRepl(False, False, True,  _('(8)')),
-    'KP_Prior':     KeyRepl(False, False, True,  _('(9)')),
-    'KP_Insert':    KeyRepl(False, False, True,  _('(0)')),
+    'KP_End':       KeyRepl(False, False, True,  _('1ᴷᴾ')),
+    'KP_Down':      KeyRepl(False, False, True,  _('2ᴷᴾ')),
+    'KP_Next':      KeyRepl(False, False, True,  _('3ᴷᴾ')),
+    'KP_Left':      KeyRepl(False, False, True,  _('4ᴷᴾ')),
+    'KP_Begin':     KeyRepl(False, False, True,  _('5ᴷᴾ')),
+    'KP_Right':     KeyRepl(False, False, True,  _('6ᴷᴾ')),
+    'KP_Home':      KeyRepl(False, False, True,  _('7ᴷᴾ')),
+    'KP_Up':        KeyRepl(False, False, True,  _('8ᴷᴾ')),
+    'KP_Prior':     KeyRepl(False, False, True,  _('9ᴷᴾ')),
+    'KP_Insert':    KeyRepl(False, False, True,  _('0ᴷᴾ')),
     'KP_Delete':    KeyRepl(False, False, True,  _('(.)')),
     'KP_Add':       KeyRepl(False, False, True,  _('(+)')),
     'KP_Subtract':  KeyRepl(False, False, True,  _('(-)')),
     'KP_Multiply':  KeyRepl(False, False, True,  _('(*)')),
     'KP_Divide':    KeyRepl(False, False, True,  _('(/)')),
     'KP_Enter':     KeyRepl(True,  False, False, _('⏎')),
+    'KP_1':         KeyRepl(False, False, True,  _('1ᴷᴾ')),
+    'KP_2':         KeyRepl(False, False, True,  _('2ᴷᴾ')),
+    'KP_3':         KeyRepl(False, False, True,  _('3ᴷᴾ')),
+    'KP_4':         KeyRepl(False, False, True,  _('4ᴷᴾ')),
+    'KP_5':         KeyRepl(False, False, True,  _('5ᴷᴾ')),
+    'KP_6':         KeyRepl(False, False, True,  _('6ᴷᴾ')),
+    'KP_7':         KeyRepl(False, False, True,  _('7ᴷᴾ')),
+    'KP_8':         KeyRepl(False, False, True,  _('8ᴷᴾ')),
+    'KP_9':         KeyRepl(False, False, True,  _('9ᴷᴾ')),
+    'KP_0':         KeyRepl(False, False, True,  _('0ᴷᴾ')),
     'Num_Lock':     KeyRepl(False, True,  True,  _('NumLck')),
     'Scroll_Lock':  KeyRepl(False, True,  True,  _('ScrLck')),
     'Pause':        KeyRepl(False, True,  True,  _('Pause')),
@@ -197,9 +208,9 @@ class LabelManager:
         self.stop()
         compose = (self.key_mode == 'composed')
         translate = (self.key_mode in ['composed', 'translated'])
-        self.kl = InputListener(self.key_press, self.btn_press,
-                                InputType.keyboard | InputType.button, compose,
-                                translate)
+        self.kl = InputListener(self.event_handler,
+                                InputType.keyboard | InputType.button,
+                                compose, translate)
         self.kl.start()
         self.logger.debug("Thread started.")
 
@@ -315,11 +326,21 @@ class LabelManager:
         self.update_text(True)
 
 
-    def key_press(self, event):
+    def event_handler(self, event):
         if event is None:
             self.logger.debug("inputlistener failure: {}".format(str(self.kl.error)))
             self.label_listener(None, None)
             return
+
+        if isinstance(event, inputlistener.KeyData):
+            self.key_press(event)
+        elif isinstance(event, inputlistener.ButtonData):
+            self.btn_press(event)
+        else:
+            self.logger.error("unhandled event type {}".format(type(event)))
+
+
+    def key_press(self, event):
         if event.symbol is None:
             # TODO: Investigate what causes this to happen.
             # I caught it once in pdb, but in this function, not in inputlistener,
@@ -513,17 +534,33 @@ class LabelManager:
         self.data.append(KeyData(datetime.now(), True, True, True, True, value))
         return True
 
-
+    # callback for mouse button presses
     def btn_press(self, event):
         if not self.enabled:
             return False
-
         if event.pressed:
             action = "pressed"
         else:
             action = "released"
         self.logger.debug("Mouse button %d %s" % (event.btn, action))
 
-        self.image_listener(
-            ButtonData(datetime.now(), event.btn, event.pressed)
-        )
+        # possible event.btn values:
+        # 1 = LMB, 2 = MMB (wheel click), 3 = RMB, 4/5 = wheel up/down,
+        # 6/7 = wheel left/right, 8+ extra buttons (e. g. thumb buttons)
+        if event.btn > 7:
+            # image_listener can only handle common buttons, redirect thumb buttons etc. to label
+            if not event.pressed:
+                return False
+            # what we refer to as "Mouse 4" has an internal value of 8,
+            # so we subtract 4 from the btn value
+            markup = GLib.markup_escape_text("M{}".format(event.btn - 4))
+
+            # show as label, treated the same as keyboard button presses
+            self.data.append(KeyData(datetime.now(), False, True,
+                         True, True, markup))
+            self.update_text()
+        else:
+            # show event in image
+            self.image_listener(
+                ButtonData(datetime.now(), event.btn, event.pressed)
+            )
